#include "SceneGraph.h"
#include "KdTree.h"
#include "X3dExporter.h"
#include "Environment.h"
#include "Polygon3.h"
#include "VssRay.h"
#include "RssTree.h"
#include "ViewCellsManager.h"
#include "RenderSimulator.h"
#include "GlRenderer.h"
#include "SamplingStrategy.h"
#include "PreprocessorThread.h"
#include "CombinedPreprocessor.h"
#include "RayCaster.h"

#include "PerfTimer.h"


namespace GtpVisibilityPreprocessor {

  extern PerfTimer viewCellCastTimer;
  extern PerfTimer pvsTimer;
  extern PerfTimer objTimer;

const bool pruneInvalidRays = true;

  
  CombinedPreprocessor::CombinedPreprocessor():
	Preprocessor()
  {
	// this should increase coherence of the samples
// 	Environment::GetSingleton()->GetIntValue("RssPreprocessor.vssSamples", mTotalSamples);
// 	Environment::GetSingleton()->GetIntValue("RssPreprocessor.samplesPerPass", mSamplesPerPass);
// 	Environment::GetSingleton()->GetIntValue("RssPreprocessor.vssSamplesPerPass",
// 											 mSamplesPerEvaluation);
	
	
  }
  

bool
CombinedPreprocessor::ComputeVisibility()
{
  
  Debug << "Preprocessor type: combined" << endl;
  cout<<"Combined Preprocessor started\n"<<flush;
  
  // if not already loaded, construct view cells from file
  if (!mLoadViewCells) {
	// construct view cells using it's own set of samples
	mViewCellsManager->Construct(this);
	
	//-- several visualizations and statistics
	Debug << "view cells construction finished: " << endl;
	mViewCellsManager->PrintStatistics(Debug);
  }
  
  
  // now decode distribution string
  char buff[1024];
  Environment::GetSingleton()->GetStringValue("RssPreprocessor.distributions",
											  buff);
  
  
  mMixtureDistribution = new MixtureDistribution(*this);
  mMixtureDistribution->Construct(buff);
  
  SimpleRayContainer rays;
  VssRayContainer vssRays;

  int lastEvaluation = 0;
  int totalVssRays = 0;
  mPass = 0;
  
  vector<VssRayContainer> rayBuffer;
  rays.reserve(mSamplesPerPass);
  rayBuffer.reserve(2*mSamplesPerPass);
  
  long startTime = GetTime();
  for (int i=0; i < mTotalSamples; i += mSamplesPerPass, mPass++) {
	mRayCaster->InitPass();
	
	cout<<"Progress : "<<(100.0f*i)/mTotalSamples<<"%"<<endl;
	rays.clear();
	vssRays.clear();
	
	
	mMixtureDistribution->GenerateSamples(mSamplesPerPass, rays);

	bool doubleRays = true;
	CastRays(rays, vssRays, doubleRays, pruneInvalidRays);
	totalVssRays += (int)vssRays.size();

	mMixtureDistribution->ComputeContributions(vssRays);

	cout<<"view cell cast time:"<<viewCellCastTimer.TotalTime()<<" s"<<endl;
	cout<<"pvs time:"<<pvsTimer.TotalTime()<<" s"<<endl;
	cout<<"obj time:"<<objTimer.TotalTime()<<" s"<<endl;

	if (mExportRays) {
	  Debug<<"Exporting rays..."<<endl<<flush;
	  char filename[256];

	  VssRayContainer contributingRays;
	  vssRays.GetContributingRays(contributingRays, 0);
	  sprintf(filename, "rss-crays-%04d.x3d", mPass);
	  ExportRays(filename, contributingRays, mExportNumRays);

	  sprintf(filename, "rss-rays-i%04d.x3d", mPass);
	  ExportRays(filename, vssRays, mExportNumRays);


	  if (mExportAnimation) {
		rayBuffer.push_back(VssRayContainer());
		vssRays.SelectRays(mExportNumRays, rayBuffer[rayBuffer.size()-1], true);
	  }

	  Debug<<"done."<<endl<<flush;
	}

	mMixtureDistribution->UpdateDistributions(vssRays);
	
	if (i - lastEvaluation >= mSamplesPerEvaluation) {
	  mViewCellsManager->PrintPvsStatistics(mStats);
	  mStats <<
		"#Pass\n" <<mPass<<endl<<
		"#Time\n" << TimeDiff(startTime, GetTime())<<endl<<
		"#TotalSamples\n" <<i<<endl<<
		"#RssSamples\n" <<totalVssRays<<endl;
	  lastEvaluation = i;

	  if (renderer) {
		ComputeRenderError();
	  }
	  
	}

#if 0
	cerr<<"deleting rays"<<endl;
	//clear all unreferenced rays
	for (int j=0; j < vssRays.size(); j++) {
	  if (vssRays[j]->RefCount() == 0)
		delete vssRays[j];
	}
	cerr<<"done."<<endl;
#endif  
	
	//	if (!mMixtureDistribution->RequiresRays())
	//	  CLEAR_CONTAINER(vssRays);

	
  }
  
  EvalViewCellHistogram();

  if (mExportAnimation && mExportRays) {
	char filename[64];
	sprintf(filename, "rss-rays-i.x3d");
	//rayBuffer.resize(mPass);
	ExportRayAnimation(filename, rayBuffer);
  }

  return true;
}

}
